/* $Id: 542x.c,v 1.13 1999/03/08 16:41:38 martin Exp $
******************************************************************************

   KGI - Cirrus Logic 542x chipset driver

   Copyright (C) 1997 Martin Erhardsed	[mee@daimi.aau.dk]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

   This code was derived from the following source of information:
   [1]     Xfree 86 vga driver
   [2]     Zen of Graphics Programming (Michael Abrash)

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER  Martin_Erhardsen
#define DRIVER_VERSION	"$Revision: 1.13 $"

#undef 	CURSOR_HW 
#define TEXT_A000

/* Set this to the address of linear addressing window if using VL-BUS */
/* #define CIRRUS_LINEAR 0xE00000 */
#define CIRRUS_LINEAR 0

/*\
|*|     GGI includes
\*/

#include <kgi/kgi.h>
#include <kgi/module.h>
#include <kgi/kgi_commands.h>
#define CIRRUS_STRINGS
#include "Cirrus/542x.h"

#include "cl542xio.inc"

#define min(a,b) (((a)<(b))?(a):(b)) 
#define max(a,b) (((a)>(b))?(a):(b))
extern struct kgi_clock clock;
extern struct kgi_ramdac ramdac;

static struct kgi_chipset chipinfo =
{
    "Cirrus Logic",		/* manufacturer         */
    "542x",			/* model                */
    GM_TEXT16 |          	/* supported modes      */
    GM_1BIT | GM_8BIT | GM_15BIT | GM_16BIT | GM_24BIT,
    0,
    {2087, 1025},		/* maximum resolution   */
    NULL,			/* Private data hook 	*/
    0x100000,			/* 256k (quasi-linear) addressable RAM size */
    { 230934, 86000000}		/* DCLK limits            */
};

static struct statetable	/* this table holds the default bits */
 default_state =		/* for text & graphics mode, will be */
{				/* filled in during mode_init        */
    0x2F,			/* MISC */
    0x00,			/* HDR  */
    {
	0x03, 0x00, 0x0F, 0x00, 0x02, 0x00, 0x12, 0x00,	/* SEQ */
	0x80, 0x0C, 0x11, 0x66, 0x5B, 0x45, 0x7E, 0x14,

	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xD8, 0x30,
	0x00, 0x01, 0x00, 0x3B, 0x2F, 0x30, 0x33, 0x1C
    },
    {
	0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x10,		/* CRT */
	0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,

	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x1x */
	0xFF, 0x00, 0x02, 0x82
    },
    {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* GRC */
	0xFF, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00
    },
    {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,		/* ATC */
	0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
	0x00, 0x00, 0x0F, 0x00, 0x00
    }
};

static unsigned long fb_text_len;
static unsigned long fb_text;

static struct io_region vga_io = {
	IO_NULL,
	0x3C0,		/* VGA IO Region Start */
	0x20,		/* 32 register (0x3C0-0x3DF) */
	0x3FF,		/* Only 10 address lines are decoded.. */
	"Cirrus Logic 542x"
};

static struct mem_region vga_graph = {
	MEM_NULL,
	0xA0000,	/* VGA Mem Region Start */
	0x20000,	/* Graphics (64K) region */
	-1,		/* Decode all bits */
	"Cirrus Logic 542x"
};

static struct mem_region cirrus_linear = {
	MEM_NULL,
	0xE00000,	/* VGA Mem Region Start */
	0x200000,	/* Graphics (64K) region */
	-1,		/* Decode all bits */
	"Cirrus Logic 542x"
};


/****************************
 * Set the page to map      *
 * (and clear old mappings) *
 ****************************/

/* 16 page grannularity for 5426+ */

static void mmap_setpage_paged_16k(struct kgi_mmio_region *mm,kgi_addr offset)
{
	int page=(offset>>17)&0x7;
	outbGRC(page << 3, 0x09),
	mm->offset=(page << 17);
}

/* 4k grannularity for older cards */

static void mmap_setpage_paged_4k(struct kgi_mmio_region *mm,kgi_addr offset)
{
	int page=(offset>>17)&0x7;
	outbGRC(page << 5, 0x09),
	mm->offset=(page << 17);
}

#define MMAP_IBM_VGA_4 (MMAP_PRIVATE|4)
#define MM_PROT_RW (MMIO_RDWR|MMIO_SHARED)

static struct kgi_mmio_region mm_text16= {
	NULL,		/* Next */
	"text16",	/* Name */
	MMAP_FRAMEBUFFER,/* Type (Linear) */
	MM_PROT_RW,	/* Protection map */

	0x10000,	/* Total size of area */
	0x10000,	/* Physical size (128k) */
	0xA0000,	/* Physical address of the windows */
	0,		/* Logical offset */

	NULL		/* Set offset function */
};
	
static struct kgi_mmio_region mm_small= {
	NULL,		/* Next */
	"Cirrus_small", /* Name */
	MMAP_FRAMEBUFFER,/* Type (Linear) */
	MM_PROT_RW,	/* Protection map */

	0x100000,	/* Total size of area */
	0x20000,	/* Physical size (128k) */
	0xA0000,	/* Physical address of the windows */
	0,		/* Logical offset */

	mmap_setpage_paged_4k	/* Set offset function */
};

static struct kgi_mmio_region mm_linear= {
	NULL,		/* Next */
	"Cirrus_linear",/* Name */
	MMAP_FRAMEBUFFER,/* Type (Linear) */
	MM_PROT_RW,	/* Protection map */

	0x100000,	/* Total size of area */
	0x100000,	/* Physical size (64k) */
	0xE00000,	/* Physical address of the windows */
	0,		/* Logical offset */

	NULL		/* Set offset function */
};

static kgiu8 PCIbus,PCIdevfn;

kgiu8 chip_version;
kgiu32 lin_adr=CIRRUS_LINEAR;

void init_542x(void)
{  
  kgiu32 maxVCLK,DRAMbw,DRAMclk;
  /* Detect MCLK */
  kgiu8 MCLK,DRAM_CR;
  static const kgiu8 fixedMCLK[4] = {0x1c, 0x19, 0x17, 0x15};

  /* Copy config (set by BIOS)*/
  default_state.SEQ[0x0F]=DRAM_CR=inbSEQ(0x0F);
  default_state.SEQ[0x1F]=MCLK=inbSEQ(0x1F);

  if (chip_version>=CLGD5424_ID){ 
    MCLK&=0x3F;
  } else {
    MCLK=fixedMCLK[DRAM_CR&3];
  };

  DRAMclk = ( 14318000 * (int)MCLK )>>4;

  /* Leave some bandwidth */
  DRAMclk = (DRAMclk * 10)/12;

  /* Correct for bus width (detected by BIOS) */
  DRAMbw = DRAMclk <<((DRAM_CR>>3)&0x03);

  maxVCLK = (chip_version>=CLGD5429_ID) ? 86000000 : 80000000;
  chipinfo.dclk.max=maxVCLK;
  ramdac.normal.max=maxVCLK;
  clock.clock.range[0].max=maxVCLK;

  /* Init ramdac structure */

  if (((DRAM_CR>>3)&3)==3) {
    /* 64 bits bus width*/
    ramdac.clock[KGIGT_TEXT16].max=min(maxVCLK,(DRAMclk*10)/3);
  } else {
    /* 16 and 32 bits bus width */
    ramdac.clock[KGIGT_TEXT16].max=min(maxVCLK,(DRAMclk*10)/6);
  }

  ramdac.clock[KGIGT_1BIT].max=min(maxVCLK,DRAMbw*2);
  ramdac.clock[KGIGT_8BIT].max=min(maxVCLK,DRAMbw);
  ramdac.clock[KGIGT_15BIT].max=
    ramdac.clock[KGIGT_16BIT].max= min(
    (chip_version>=CLGD5426_ID) ? maxVCLK : maxVCLK/2,DRAMbw/2);
  ramdac.clock[KGIGT_24BIT].max=
    min((chip_version>=CLGD5436_ID) ? maxVCLK : maxVCLK/3,DRAMbw/3);
  if ((chip_version>=CLGD5434_ID)&&(((DRAM_CR>>3)&3)==3))
    {
      ramdac.clock[KGIGT_32BIT].max=min(maxVCLK,DRAMbw/4);
      chipinfo.modes |= GM_32BIT;
    }
  ramdac.modes=chipinfo.modes;
}  


int detect_542x(void)
{
	int lock,memory_size,cirrus_bus;

	int Status = 0, i, j;

	NOTICE("Looking for PCI cards");

	if (pcibios_present()) {
	  
		/*	First, figure out if there is a CL54xx on the PCI bus.
		 */
		for (i = 0; PCIDevice_id[i]; i++) {
			
			j = 0;	/* Start with the first card */
			
			Status = pcibios_find_device(PCI_VENDOR_ID_CIRRUS,
						     PCIDevice_id[i], j,
						     &PCIbus,
						     &PCIdevfn);
			j++;
			
			if (Status == PCIBIOS_SUCCESSFUL) {
				chip_version=PCIDevice_id[i];
				break;
			}
			
		}
		
		if (Status != PCIBIOS_SUCCESSFUL) goto no_pci;

		strcpy(chipinfo.model,chip_strings[(PCIDevice_id[i]>>2)-0x22]);
		DEBUG1("Found %s on bus: %d, device ID: %d\n",
		       chipinfo.model,PCIbus,PCIdevfn);

		/*	Enable access to memory and io ports */

		pcibios_write_config_word(PCIbus, PCIdevfn,PCI_COMMAND, 
					  PCI_COMMAND_IO | PCI_COMMAND_MEMORY);
  
		/*	Determine Linear framebuffer access. */

		pcibios_read_config_dword(PCIbus, PCIdevfn,
					  PCI_BASE_ADDRESS_0, &lin_adr);
		
		cirrus_linear.base = (lin_adr &= 0xFF000000);
		cirrus_linear.size = 0x01000000;   /* 16M */ 
		
		DEBUG1("Linear Frame buffer address: %x\n", lin_adr);
		
		if (lin_adr && mem_check_region(&cirrus_linear)) {
			
			ERROR("CL542x linear frame buffer busy");
			lin_adr=0;
			return -E(CHIPSET, UNKNOWN);
		}

	}
no_pci:
	outbMISC(inb(MISCr));	/* adjust IO base       */


        outbSEQ(0x12,0x06);
	lock = inbSEQ(0x06);
	if (lock!=0x12)		/* First Check for Cirrus Chipset */
	{
		ERROR("");
		return -E(CHIPSET,UNKNOWN);
	}

	chip_version = inbCRT(0x27);

	if (chip_version<CLGD5420_ID)   return	-E(CHIPSET,UNKNOWN);

	if (chip_version<CLGD5430_ID) {
		memory_size = 256*1024 << ( ( inbSEQ(0x0A) >> 3 ) & 3 );
	} else {
		memory_size = 256*1024 << ( inbSEQ(0x15) & 7 );
	}
	
	chipinfo.ramsize = memory_size;
	mm_linear.size   = memory_size;
	mm_linear.psize  = memory_size;

	switch ((inbSEQ(0x17) >> 3) & 0x07) 
	{
	case	2:
	case	6:	cirrus_bus = CIRRUS_BUS_VLB ;	break ;
	case	4:	cirrus_bus = CIRRUS_BUS_PCI ;   break ;
	case	7:	cirrus_bus = CIRRUS_BUS_ISA ;   break ;
	default:	cirrus_bus = CIRRUS_BUS_UNKNOWN ;
	}

	NOTICE ( "Cirrus Logic %s, %s bus %dK Detected\n", 
		 chip_strings [ (chip_version>>2) - 0x22 ], 
		 bus_strings [ cirrus_bus ], memory_size>>10 ) ;

	/* BUGBUG
	 * We should actually try to detect, 
	 * instead of just _assuming_
	 */

	fb_text =  0x000A0000;
	fb_text_len = 0x10000;	/* 64k framebuffer */

	return EOK;
}

/*\
|*|  GGI interface
\*/

/*
 * Mode specific frame buffers
 */

#include "542x_text16.inc"

#include "cl542xfb.inc"

#ifdef DEBUG
static void printk_mode (struct kgi_mode *mode, char *msg)
{
	#define M(X)	mode->X
	#define R(X)	mode->request.X
	#define T(X)	M(X.width), M(X.blankstart), M(X.syncstart),\
			M(X.syncend), M(X.blankend), M(X.total), M(X.polarity)

	NOTICE("%s\n",msg);
	NOTICE("mode = {\n");
        NOTICE("   { %i,  /* frames */\n");
        NOTICE("   { %i, %i},  /* vis	ible */\n",
	       R(frames), R(visible.x),R(visible.y));
	NOTICE("   { %i, %i},   /* virtual */\n",
                          R(virt.x),R(virt.y));
	NOTICE("   { %i, %i},   /* size [mm] */\n",
                          R(size.x),R(size.y));
	NOTICE("   0x%x,\n", R(graphtype));
	NOTICE("   { %i, %i}},   /* textgrid */\n",
                          R(dpp.x),R(dpp.y));
	NOTICE("   { %i, %i, %i, %i, %i, %i, %i}, /* x timing */\n",
                          T(x));
	NOTICE("   { %i, %i, %i, %i, %i, %i, %i}, /* y timing */\n",
                          T(y));
	NOTICE("   { %i, %i},   /* magnify */\n",
                          M(magnify.x), M(magnify.y));
	NOTICE("   %x,          /* flags   */\n", M(flags));
	NOTICE("   %i,          /* dclk    */\n", M(dclk));
	NOTICE("   { %i, %i},   /* clk     */\n", 
                          M(clk.mul), M(clk.div));
	NOTICE("   { %i, %i}    /* lclk    */\n"
	       KERN_DEBUG "}\n",
                          M(lclk.mul), M(lclk.div));

	#undef M
	#undef R
	#undef T
}
#endif /* #ifdef DEBUG */

static void chipset_setup_dpy(struct kgi_display *dpy,struct kgi_mode *mode)
{
  switch (mode->request.graphtype) {
    case KGIGT_TEXT16:
      dpy->set_font=text16_set_font;
      dpy->set_origin=text16_set_origin;
      dpy->set_split=text16_set_split;
      dpy->set_frame=text16_set_frame;
      dpy->show_cursor=text16_show_cursor;
      dpy->hide_cursor=text16_hide_cursor;
      dpy->undo_cursor=text16_undo_cursor;
      dpy->show_pointer=text16_show_pointer;
      dpy->hide_pointer=text16_hide_pointer;
      dpy->undo_pointer=text16_undo_pointer;
      dpy->fb.dpy=dpy;
      dpy->fb.size.x = mode->request.virt.x;
      dpy->fb.size.y = mode->request.virt.y;
      dpy->fb.bpp    = MAP_16_BPP;
      dpy->fb.pbuf   = phys_to_virt(fb_text);
      dpy->fb.zbuf   = NULL;
      dpy->fb.ops    = &text16_color_io;
      KGIM_PRIVATE(dpy, driver)=text16_data_realloc(KGIM_PRIVATE(dpy,driver));
      if (KGIM_PRIVATE(dpy, driver) == NULL)
              INTERNAL_ERROR;
      break;
    default:
      dpy->set_font=NULL;
      dpy->set_origin=cirrus_set_origin;
      dpy->set_split=cirrus_set_split;
      dpy->set_frame=NULL;
      dpy->show_cursor=NULL;
      dpy->hide_cursor=NULL;
      dpy->undo_cursor=NULL;
      dpy->show_pointer=NULL;
      dpy->hide_pointer=NULL;
      dpy->undo_pointer=NULL;
      dpy->fb.dpy = dpy;
      dpy->fb.size.x = 0;
      dpy->fb.size.y = 0;
      dpy->fb.bpp    = 0;
      dpy->fb.pbuf   = NULL;
      dpy->fb.zbuf   = NULL;
      dpy->fb.ops    = NULL;
      kgim_accel_set_mode(dpy,mode);
  }
  memcpy(&(dpy->mode),mode,sizeof(struct kgi_mode));
}

kgi_sint kgim_chipset_init(struct kgi_card *card, struct kgi_display *dpy)
{
    int err;

    NOTICE("Cirrus Logic 542x Chipset Driver " DRIVER_VERSION "\n");

    if (detect_542x()==EOK) {
      //    	if ((err=io_check_region(&vga_io)))
      //		return err;
    	if ((err=mem_check_region(&vga_graph)))
		return err;
    	if ((err=mem_check_region(&cirrus_linear)))
		return err;

    	io_claim_region(&vga_io);
    	mem_claim_region(&vga_graph);

	init_542x();  /* Set DCLK limits */

    	if (lin_adr) mem_claim_region(&cirrus_linear);

	text16_data_init();
	
	strcpy(card->manufact,chipinfo.manufact);
	strcpy(card->model,chipinfo.model);
	card->modes=chipinfo.modes;
	card->flags=chipinfo.flags;
	card->maxdots=chipinfo.maxdots;

	card->chipset=&chipinfo;

	return err;
    } 

    return -E(CHIPSET,FAILED);
}

void kgim_chipset_done(struct kgi_display *dpy)
{
  kgim_chipset_set_mode(dpy,&(dpy->card->preset));

  io_free_region(&vga_io);
  mem_free_region(&vga_graph);
  if (lin_adr) mem_free_region(&cirrus_linear);

}

static void adjust_timing(struct kgi_mode *mode)
{
    mode->x.width      &= -8;
    mode->x.blankstart &= -8;
    mode->x.syncstart  &= -8;
    mode->x.syncend    &= -8;
    mode->x.blankend   &= -8;
    mode->x.total = (mode->x.total+7) & ~7;

    if ((mode->x.syncend & 0xF8) == (mode->x.syncstart & 0xF8))
	mode->x.syncend -= 8;
 
    if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
	mode->y.syncend--;

    if (chip_version<=CLGD5424_ID) {
    if ((mode->x.blankend & 0x1F8) == (mode->x.blankstart & 0x1F8))
	mode->x.blankend -= 8;
 
    if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
	mode->y.blankend--;
    }
}

kgi_sint kgim_chipset_check_mode(struct kgi_display *dpy,
				 struct kgi_mode *mode,
				 enum kgi_tm_cmd *cmd)
{
    int bpp = 0;
    int textx,texty;
    textx = mode->request.dpp.x;
    texty = mode->request.dpp.y;

    switch (mode->request.graphtype) 
    {
        case KGIGT_32BIT:
	  if (chip_version<CLGD5430_ID) return -E(CHIPSET,BOGUS);
	  bpp = 32; break;
        case KGIGT_24BIT:
	  if (chip_version<CLGD5436_ID) mode->clk.mul=3;
	  bpp = 24; break;
	case KGIGT_15BIT:
        case KGIGT_16BIT:
	  if (chip_version<CLGD5426_ID) mode->clk.mul=2;
	  bpp = 16; break;
	case KGIGT_8BIT:
	  bpp =  8;break; 	       
        case KGIGT_1BIT:
	  bpp =  1; break;	
	case KGIGT_TEXT16:
	  bpp = 16; break;
	default:
	  DEBUG2("%s: graphtype %i not supported.\n",
		 __FILE__, mode->request.graphtype);
	  return -E(CHIPSET,FAILED);
    }

#define XVIRTUAL (mode->request.virt.x)
#define YVIRTUAL (mode->request.virt.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

    switch (*cmd) {
	case CMD_PROPOSE:
	    if (mode->request.graphtype == KGIGT_TEXT16) {

		mode->magnify.x = mode->magnify.y = 1;
		if (textx != 8) {
		    DEBUG2("%s: invalid char width (%i).\n",__FILE__,textx);
		    return -EINVAL;
		}

		if (XVIRTUAL==0 && YVIRTUAL==0) {
		  XVIRTUAL = XVISIBLE;
		  YVIRTUAL = YVISIBLE;
		}

		if (XVIRTUAL==0) 
		  XVIRTUAL = fb_text_len / (YVIRTUAL * 2);

		if (YVIRTUAL==0) 
		  YVIRTUAL = fb_text_len / (XVIRTUAL * 2);

		if (XVIRTUAL < XVISIBLE) XVIRTUAL = XVISIBLE;
		if (YVIRTUAL < YVISIBLE) YVIRTUAL = YVISIBLE;

		if ( ( 2 * XVIRTUAL * YVIRTUAL ) > fb_text_len ) {
		    DEBUG1("%s: not enough text memory.\n", __FILE__);
		    return -ENOMEM;
		}
	    } else {		/* mode->request.graphtype == KGIGT_TEXT16 */

		XVISIBLE = (XVISIBLE + 7) & ~0x7;

		mode->magnify.x = 1;
		mode->magnify.y = (350+YVISIBLE-1)/YVISIBLE;

		if (mode->magnify.y > 16) return -EINVAL;

		if (XVIRTUAL==0 && YVIRTUAL==0) {
		  XVIRTUAL = XVISIBLE;
		  YVIRTUAL = YVISIBLE;
		}

		if (XVIRTUAL==0) 
		  XVIRTUAL = ((chipinfo.ramsize*8)/(YVIRTUAL*bpp))&~0xF;

		if (YVIRTUAL==0) 
		  YVIRTUAL = (chipinfo.ramsize*8)/(XVIRTUAL*bpp+168);

		if (XVIRTUAL < XVISIBLE) XVIRTUAL = XVISIBLE;
		if (YVIRTUAL < YVISIBLE) YVIRTUAL = YVISIBLE;


		switch(mode->request.graphtype)
		  {
		  case KGIGT_1BIT:
		    XVIRTUAL=(XVIRTUAL+0xF)&~0xF; break;
		  case KGIGT_8BIT:
		  case KGIGT_24BIT:
		    XVIRTUAL=(XVIRTUAL+0x7)&~0x7; break;
		  case KGIGT_15BIT:
		  case KGIGT_16BIT:
		    XVIRTUAL=(XVIRTUAL+0x3)&~0x3; break;
		  case KGIGT_32BIT:
		    XVIRTUAL=(XVIRTUAL+0x1)&~0x1; break;
		  default:
		  }

		if ((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize) {
		   if ((XVISIBLE * YVISIBLE * bpp / 8) > chipinfo.ramsize) {
		     DEBUG1("%s: not enough memory to"
			   " do virtual %ix%i at %i bpp\n",
			    __FILE__, XVIRTUAL, YVIRTUAL, bpp);
		   return -ENOMEM;
                  } 
                  XVIRTUAL=XVISIBLE;
                  YVIRTUAL=YVISIBLE;
		}
	    }

	    mode->x.width = mode->request.visible.x * mode->magnify.x * mode->request.dpp.x;
	    mode->y.width = mode->request.visible.y * mode->magnify.y * mode->request.dpp.y;

	    return EOK;

	case CMD_LOWER:

	    if (mode->dclk < chipinfo.dclk.min) {
		DEBUG2("%s: DCLK too low (%i).\n", __FILE__, mode->dclk);
		return -E(CHIPSET,FAILED);
	    }
	    mode->dclk = (mode->dclk > chipinfo.dclk.max)
		? chipinfo.dclk.max : mode->dclk;

	    adjust_timing(mode);
	    return EOK;

	case CMD_RAISE:

	    if (mode->dclk > chipinfo.dclk.max) {
		DEBUG1("%s: DCLK too high (%i).\n",__FILE__, mode->dclk);
		return -E(CHIPSET,FAILED);
	    }
	    mode->dclk = (mode->dclk < chipinfo.dclk.min)
		? chipinfo.dclk.min : mode->dclk;

	    adjust_timing(mode);
	    return EOK;

	case CMD_CHECK:

    	    if (mode->lclk.mul != 1) {
	    	DEBUG1("%s: invalid lclk multiplier %d.\n",\
                             __FILE__,mode->lclk.mul);
	    	return -E(CHIPSET,FAILED);
	    }

	    switch (mode->lclk.div) {
	    	case 1:
	    	    break;
	    	default:
		  DEBUG1("%s: invalid CLK to LCLK divider %d.\n",\
			       __FILE__,mode->lclk.div);
		return -E(CHIPSET,FAILED);
	    }
	    if (mode->request.graphtype == KGIGT_TEXT16) {
		mode->flags |= FL_CHIP_COLORTEXT;
		if ((mode->magnify.x != 1) || (mode->magnify.y != 1)
		    || (textx != 8) || (texty < 1) || (texty > 32)
		    || ((2 * XVIRTUAL * YVIRTUAL) > fb_text_len)
		    || (mode->dclk > chipinfo.dclk.max)
		    || (mode->dclk < chipinfo.dclk.min)) {
		    DEBUG1("%s: hw check failed\n", __FILE__);
		    return -E(CHIPSET,FAILED);
		}
	    } else {

		if (((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize)
		    || (mode->magnify.x != 1 ) || (mode->magnify.y > 16)
		    || /* (mode->clk.mul != 1) || */ (mode->clk.div != 1)
		    || (mode->dclk < chipinfo.dclk.min)
		    || (chipinfo.dclk.max < mode->dclk))
		    return -E(CHIPSET,FAILED);
	    }

#define TM(X) (((unsigned) chipinfo.maxdots.x < mode->x.X)\
		       || ((unsigned) chipinfo.maxdots.y < mode->y.X))
	    if (TM(width) || TM(blankstart) || TM(syncstart)
		|| TM(syncend) || TM(blankend) || TM(total)) {
		TRACE(NOTICE("%s: timing check failed\n", __FILE__));
		TRACE(NOTICE("%s: Maxdots are (%d,%d)\n", __FILE__,
			     (unsigned)chipinfo.maxdots.x,
			     (unsigned)chipinfo.maxdots.y));
		return -E(CHIPSET,FAILED);
	    }
#undef TM

	    /*\
	    |*|        Setup mmap 
            \*/

	      if (mode->request.graphtype==KGIGT_TEXT16) {
		mm_text16.next=mode->mmio;
		mode->mmio=&mm_text16;
	      } else if (lin_adr) {
		mm_linear.next=mode->mmio;
		mm_linear.pstart=lin_adr;
		mode->mmio=&mm_linear;
	      } else {
		if (chip_version>=CLGD5426_ID)
		  mm_small.set_offset=mmap_setpage_paged_16k; 
		mm_small.next=mode->mmio;
		mode->mmio=&mm_small;
	      }
	      mode->fb=ioremap(mode->mmio->pstart,mode->mmio->psize);
	      mode->fb_size=mode->mmio->psize;
	      mode->fsize= (mode->request.virt.x*
			    mode->request.virt.y*8)/bpp;
	    return EOK;

	default:
	    INTERNAL_ERROR;
	    return -E(CHIPSET,FAILED);
    }

#undef XVIRTUAL
#undef YVIRTUAL
#undef XVISIBLE
#undef YVISIBLE
}

void kgim_chipset_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
    struct statetable setup = default_state;
    register int foo;

    /*\
    |*|        Setup bits per pixel and logical screen width
    \*/

    switch (mode->request.graphtype) {
	case KGIGT_TEXT16:
	    foo = mode->request.virt.x >> 1;		
	    /* due to the addressing mode used. */
	    break;
	case KGIGT_1BIT:
	    foo = mode->request.virt.x >> 4;
	    break;
	case KGIGT_8BIT:
	    foo = (mode->request.virt.x>>3)*1;
	    break;
        case KGIGT_15BIT:
        case KGIGT_16BIT:
	    foo = (mode->request.virt.x>>3)*2;
	    break;
        case KGIGT_24BIT:
	    foo = (mode->request.virt.x>>3)*3;
	    break;
        case KGIGT_32BIT:
	    foo = (mode->request.virt.x>>3)*4;
	    break;
	default:
	    INTERNAL_ERROR;
	    return;
    }

    setup.CRT[CRT_LOGICALWIDTH] = foo;
    setup.CRT[0x1B] |= (foo&0x100) >>4;

    if (mode->request.graphtype == KGIGT_TEXT16) {

	setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
	setup.SEQ[0x02]  = 0x03;
	setup.CRT[0x09] |= mode->request.dpp.y - 1;
	setup.CRT[0x14] |= CR14_UNDERLINE_MASK;
	setup.CRT[0x17]  = CR17_CGA_ADDR_WRAP
	    | CR17_ENABLE_SYNC | CR17_CGA_BANKING
	    | CR17_HGC_BANKING;
	setup.GRC[0x05] |= GR5_ODD_EVEN;
	setup.GRC[0x06] |= GR6_CHAIN_OE | GR6_MAP_A0_64;
	setup.ATC[0x13]  = 0x00;

    } else {			/* mode->request.graphtype == KGIGT_TEXT16 */

	setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
	setup.SEQ[0x02] = SR02_PLANE_MASK;

	/* Use 16k page offset if supported */
	if (chip_version>=CLGD5426_ID) 
	  setup.GRC[0x0B]|= 0x20;	  	

	/* Enable linear adressing */
	if ((mode->request.graphtype!=KGIGT_4BIT)&&(lin_adr)) 
	  setup.SEQ[0x07]|= 0xE0;

	for (foo = 0; foo < 16; foo++)
	    setup.ATC[foo] = foo;
	setup.ATC[0x10] |= AR10_GRAPHICS;

	setup.GRC[0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_128;

	setup.CRT[0x03] |= CR03_IS_VGA;
	setup.CRT[0x09] |= (mode->magnify.y - 1) & CR09_CHARHEIGHT_MASK;
	setup.CRT[0x17] = 0xE3;

	setup.HDR=0x00;

	switch (mode->request.graphtype) {
	case KGIGT_1BIT:
	      setup.SEQ[0x02]  = 0x01;
	      setup.SEQ[0x04]  = 0x06;
	      setup.GRC[0x01]  = 0x0e;	/* Set/Reset all but plane 0 */
	      setup.GRC[0x04]  = 0x00;
	      setup.GRC[0x05] |= GR5_WRITEMODE0;
	      setup.ATC[0x12] =0x01;         /* use 1 bit */
	      break;
	case KGIGT_8BIT:
	      setup.SEQ[0x04] =0x0E;
	      setup.SEQ[0x07]|=0x01; /* ??? */
	      setup.GRC[0x05] =0x40;
	      break;
	case KGIGT_15BIT:
	      setup.SEQ[0x04] =0x0E;
	      setup.SEQ[0x07]|=0x07; /* ??? */
	      setup.GRC[0x05] =0x40;
	      setup.HDR=0xE0;
	      break;
	case KGIGT_16BIT:
	      setup.SEQ[0x04] =0x0E;
	      setup.SEQ[0x07]|=0x07; /* ??? */
	      setup.GRC[0x05] =0x40;
	      setup.HDR=0xE1;
	      break;
	case KGIGT_24BIT:
	      setup.SEQ[0x04] =0x0E;
	      setup.SEQ[0x07]|=0x05; /* ??? */
	      setup.GRC[0x05] =0x40;
	      setup.HDR=0xE5;
	      break;
	case KGIGT_32BIT:
	      setup.SEQ[0x04]=0x0E;
	      setup.SEQ[0x07]|=0x09; /* ??? */
	      setup.GRC[0x05]=0x40;
	      setup.HDR=0xE5;
	      break;
	    default:
		INTERNAL_ERROR;
		return;
	}
    }

    /*\
    |*|        setup horizontal timing
    \*/

    /* horizontal display end */
    setup.CRT[0x01] = (mode->x.width >> 3) - 1;	

    /* horizontal blank start */
    setup.CRT[0x02] = (mode->x.blankstart >> 3);	

    /* horizontal sync start  */
    setup.CRT[0x04] = (mode->x.syncstart >> 3);	

    foo = (mode->x.syncend >> 3);	/* horizontal sync end    */
    setup.CRT[0x05] |= foo & 0x1F;

    /* horizontal blank end   */
    foo = (mode->x.blankend >> 3);
    setup.CRT[0x03] |= foo & 0x1F;
    setup.CRT[0x05] |= (foo & 0x20) << 2;
    setup.CRT[0x1A] |= (foo & 0xC0) >> 2;

    /* horizontal total       */
    setup.CRT[0x00] = (mode->x.total >> 3) - 5;
        

    /*\
    |*|        setup vertical timing
    \*/

    foo = mode->y.width - 1;	/* vertical display end   */
    setup.CRT[0x12]  =  foo;
    setup.CRT[0x07] |= (foo & 0x100) >> 7;
    setup.CRT[0x07] |= (foo & 0x200) >> 3;

    foo = mode->y.blankstart - 1;	/* vertical blank start   */
    setup.CRT[0x15]  =  foo;
    setup.CRT[0x07] |= (foo & 0x100) >> 5;
    setup.CRT[0x09] |= (foo & 0x200) >> 4;

    foo = mode->y.syncstart;	/* vertical sync start    */
    setup.CRT[0x10] |=  foo;
    setup.CRT[0x07] |= (foo & 0x100) >> 6;
    setup.CRT[0x07] |= (foo & 0x200) >> 2;

    foo = mode->y.syncend;	/* vertical sync end      */
    setup.CRT[0x11] |= foo & 0x0F;

    foo = mode->y.blankend + 1;	/* end vertical blank     */
    setup.CRT[0x16]  =  foo;
    setup.CRT[0x1A] |= (foo>>2) & 0xC0;

    foo = mode->y.total - 2;	/* vertical total         */
    setup.CRT[0x06]  =  foo;
    setup.CRT[0x07] |= (foo & 0x100) >> 8;
    setup.CRT[0x07] |= (foo & 0x200) >> 4;

    setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
    setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;

    /*\
    |*|        setup clock
    \*/

      if (mode->clock == -1) INTERNAL_ERROR;
      setup.SEQ[0x0E] = mode->clock >> 8;
      setup.SEQ[0x1E] = mode->clock;

    /*\
    |*|        Ok. We'll do the programming now.
    \*/


    DEBUG3("Set mode:");

    kgim_monitor_set_mode(dpy,mode);
    kgim_ramdac_set_mode(dpy,mode);
    restore_state(&setup);

    chipset_setup_dpy(dpy,mode);

    DEBUG3("Mode set complete");

}

kgi_sint kgim_chipset_command(struct kgi_graphic *state,kgi_uint cmd,kgi_uint iobuf)
{
	kgi_coord  *coord;
	struct kgi_font *font;
	struct kgi_display *dpy;

	dpy = state->dev.dpy;
	switch (cmd) {
		case CHIP_SETDEFFONT:
		font = (struct kgi_font *)(state->io_buf);
		text16_set_font(dpy, font, 0, font->info->positions);
		return 0;
		case CHIP_SETSPLITLINE:
			if (iobuf > state->dev.mode.visible.y)
				return -E(CHIPSET,BOGUS);

			if (state->dev.org[0].y > state->dev.mode.visible.y - iobuf )
				kgi_set_origin(&(state->dev), 0, 
						state->dev.org[0].x,
						state->dev.mode.visible.y - iobuf );

			kgi_set_split(&(state->dev),iobuf*dpy->mode.magnify.y);
			return 0;
		case CHIP_SETVISFRAME:
			coord = (kgi_coord *)(state->io_buf);

			if (coord->x < 0 || 
			    coord->x > state->dev.mode.virt.x - state->dev.mode.visible.x)
				return -E(CHIPSET,BOGUS);

			if (coord->y < 0 || dpy->mode.magnify.y<1 ||
			    coord->y > state->dev.mode.virt.y - state->dev.split/dpy->mode.magnify.y)
				return -E(CHIPSET,BOGUS);

			kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
			return 0;
		default:
			break;
	}

	return -E(CHIPSET,UNKNOWN);
}
