/* $Id: et4000_fb.inc,v 1.7 1998/11/24 18:28:09 marcus Exp $
******************************************************************************

   Tseng ET4000 FB operators

   Copyright (C) 1997      Jason McMullan, [jmcc@ggi-project.org]
   Copyright (C) 1997	     Andreas Vogler, [duff@allgaeu.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

	This code was derived from the following source of information:
	[1]     IBM vga driver by Jason McMullan
	[2]     vgadocs4 by Finn Thoegersen

******************************************************************************
*/

#ifndef _ET4000_FB_INC_
#define _ET4000_FB_INC_

#ifndef MAINTAINER
# define MAINTAINER	"Gernot Ziegler, <gz@lysator.liu.se>"
#endif

static void et4000_set_origin(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	IS_CRITICAL;
	static kgi_uint last_xdot = -1;
	static kgi_uint last_ydot = -1;
	static kgi_uint old_offset = -1;
	register kgi_uint xpix=0,xdot=0;
	register kgi_uint ypix=0,ydot=0;
	register kgi_uint offset;
	kgi_uint fb_width=0;
	
	switch (dpy->mode.request.graphtype) {
/*	case KGIGT_TEXT16:
		xpix = x / dpy->mode.request.dpp.x;
		xdot = x % dpy->mode.request.dpp.x;
		ypix = y / dpy->mode.request.dpp.y;
		ydot = y % dpy->mode.request.dpp.y;
		fb_width = dpy->mode.request.virt.x/
			dpy->mode.request.dpp.x*2;
*/		
	case KGIGT_8BIT:
		xpix=x;
		xdot=0;
		ypix=y;
		ydot=0;
		fb_width = dpy->mode.request.virt.x;
		break;

	case KGIGT_15BIT:
	case KGIGT_16BIT:
		xpix=2*x;
		xdot=0;
		ypix=y;
		ydot=0;
		fb_width = dpy->mode.request.virt.x*2;
		break;	

	case KGIGT_24BIT:
		xpix=3*x;
		xdot=0;
		ypix=y;
		ydot=0;
		fb_width = dpy->mode.request.virt.x*3;
		break;	

	case KGIGT_32BIT:
		xpix=4*x;
		xdot=0;
		ypix=y;
		ydot=0;
		fb_width = dpy->mode.request.virt.x*4;
		break;	
		
	default:
		INTERNAL_ERROR;
	}
	
	offset = xpix + ypix * fb_width;
	
	ENTER_CRITICAL;
	
	if (old_offset != offset) {
		offset &= 0xFFFFF;
		outbCRT(offset >> 8, CRT_STARTADDR_H);
		outbCRT(offset,      CRT_STARTADDR_L);
		if(chip==ET4000) {
			offset &= 0x3FFFF;
			outb((inbCRT(CRT_EXT_ADDR) &
			      ~CR33_DISPLAYSTART_MASK) | (offset >> 16), CRT_EXT_ADDR);
		} else {
			outb((inbCRT(CRT_EXT_ADDR) &
			      ~CR33_DS_MASK_W32X) | (offset >> 16), CRT_EXT_ADDR);
		}
		old_offset = offset;
	}
	
	if (last_ydot != ydot) {
		outbCRT(ydot & CR08_ROWSCAN_MASK, CRT_PRESETROW);
		last_ydot = ydot;
	}
	
	if (last_xdot != xdot) {
		
		register kgi_uint hpan;
		
		if (dpy->mode.request.graphtype==KGIGT_TEXT16 &&
		    dpy->mode.request.dpp.x == 9) {
			
			hpan = (xdot == 0)
				? 0x08 : (xdot-1) & AR13_DATA_MASK;
		} else {
			
			hpan = xdot & AR13_DATA_MASK;
		}
		WAIT_HVRETRACE();
		outbATC(hpan, ATC_HPAN);
		outb( ATCI_ENABLEDISPLAY, ATCI);
		last_xdot = xdot;
	}
	
	LEAVE_CRITICAL;
}

#ifdef NOT_USED_CODE
static void et4000_set_split(struct kgi_display *dpy, int line)
{
	register kgiu8 foo;

	DEBUG2("set_split called with %i\n", line);

	line *= dpy->mode.magnify.y;
	
	outbCRT(line & 0xFF, CRT_LINECOMPARE);
	
	foo = (line & 0x0100)
		? inbCRT(0x07) | CR07_LINECOMPARE_B8
		: inbCRT(0x07) & ~CR07_LINECOMPARE_B8;
	outbCRT(foo, 0x07);

	foo = (line & 0x0200)
		? inbCRT(0x09) | CR09_LINECOMPARE_B9
		: inbCRT(0x09) & ~CR09_LINECOMPARE_B9;
	outbCRT(foo, 0x09);

	foo = (line & 0x0400)
		? inbCRT(0x35) | CR35_LINECOMPARE_B10
		: inbCRT(0x35) & ~CR35_LINECOMPARE_B10;
	outbCRT(foo, 0x35);
}

#endif /* NOT_USED_CODE */

#define	et4000_set_frame	NULL	/* !!! implement! */

/*
 * support for the W32 hardware cursor
 */

/* initialize an area in video memory to be used as
 * hw pointer
 */

void w32_init_hw_pointer(struct kgi_display * dpy)
{
#if 0
	Sprite_out8(0 & 0xFF, 0xE8);
	Sprite_out8((0 >> 8)& 0xFF, 0xE9);
	Sprite_out8((0 >> 16) & 0xFF, 0xEA);

	if(chip==ET4000_W32p) {
		Sprite_out8(2, 0xEB);
	} else {
		Sprite_out8(4, 0xEB);
		Sprite_out8(0, 0xEC);
	}

	Sprite_out8((Sprite_in8(0xEF) & ~0x07) | 0x02, 0xEF); 
#endif
}

void w32_show_pointer(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	Sprite_out8(x & 0xFF, 0xE0);
	Sprite_out8((x >> 8) & 0x0F, 0xE1);
	Sprite_out8(y & 0xFF, 0xE4);
	Sprite_out8((y >> 8) & 0x0F, 0xE5);

	Sprite_out8((Sprite_in8(0xF7) | 0x80), 0xF7); 
}

void w32_hide_pointer(struct kgi_display *dpy)
{
	Sprite_out8((Sprite_in8(0xF7) & ~0x80), 0xF7); 
}

/* we don't need w32_undo_pointer for a hardware pointer */

#endif /* _ET4000_FB_INC */

